
require 'fileutils'
require 'kyotocabinet'
#require 'tomte/structures/persistent_store'

require 'ygg/provisioner/task'
require 'ygg/provisioner/tools'
require 'ygg/dns/authoritative_provisioner/zone'

module Ygg
module Dns
module AuthoritativeProvisioner

class Task < Ygg::Provisioner::Task
  class ZoneNotFound < PermanentFailure ; end

  def find_zone(id)
    Zone.find(id)
  rescue Zone::NotFound
    raise ZoneNotFound
  end

  def hash_deep_workaround(hash)
    hash.delete '_type'
    hash.delete :_type
    hash.each do |k,v|
      hash_deep_workaround(v) if v.kind_of?(Hash)
      if v.kind_of?(Array)
        v.each do |x| hash
          hash_deep_workaround(x) if x.kind_of?(Hash)
        end
      end
    end
  end

  operation :list do
    help_caption 'list all hostings'
    help_text ''
  end

  def list
    res = nil
    db_path = agentcfg[:zones_db_path]
    db = KyotoCabinet::DB::new
    begin
      if db.open(db_path, KyotoCabinet::DB::OWRITER | KyotoCabinet::DB::OCREATE)
        res = []
        db.each { |k,v| res << [ k, Marshal.load(v)[:cfg][:name] ] }
        res = Hash[res]
      end
    rescue
      raise DatabaseError.new("Cannot open database #{db_path}: #{db.error}")
    ensure
      unless db.close
        raise DatabaseError.new("Cannot close database #{db_path}: #{db.error}")
      end
    end

    res
  end

  operation :info do
    help_caption 'get informations from zone'
    help_text <<-EOT.unindent
      parameters:
        id:
    EOT
  end
  def info(id)
    zone = nil

    db_path = agentcfg[:zones_db_path]
    db = KyotoCabinet::DB::new
    begin
      if db.open(db_path, KyotoCabinet::DB::OWRITER | KyotoCabinet::DB::OCREATE)
        zone = db.get(id)
      end
    rescue
      raise DatabaseError.new("Cannot open database #{db_path}: #{db.error}")
    ensure
      unless db.close
        raise DatabaseError.new("Cannot close database #{db_path}: #{db.error}")
      end
    end

    raise ZoneNotFound if !zone
    zone = Marshal.load(zone)

    zone = zone.marshal_dump if zone.kind_of?(OpenStruct)

    hash_deep_workaround(zone)

    zone
  end

  operation :create do
    help_caption 'create'
    help_text <<-EOT.unindent
      parameters:
        request:    The whole request
    EOT
  end
  def create
    res = nil
    zone = Zone.new(request_data)

    log.info "######### CREATING #{zone.name} ############"
    run_event_script zone, :before_create

    zone.logging_on(plog) do
      zone.save!
      zone.create!
    end

    run_event_script zone, :after_create
    log.info "######### CREATING #{zone.name} DONE #######"

    res
  end

  operation :update do
    help_caption 'update'
    help_text <<-EOT.unindent
      parameters:
        request:    The whole request
    EOT
  end

  def update
    begin
      zone = find_zone(request_data[:id])
      zone.update_config(request_data)
    rescue
      # In case of corrupted local database
      zone = Zone.new(request_data)
    end

    log.info "######### UPDATING #{zone.name} ############"
    run_event_script zone, :before_update

    res = nil
    zone.logging_on(plog) do
      zone.save!
      zone.update!
    end

    run_event_script zone, :after_update
    log.info "######### UPDATING #{zone.name} DONE #######"

    res
  end

  operation :delete do
    help_caption 'deprovision the zone'
    help_text <<-EOT.unindent
      parameters:
        id
    EOT
  end
  
  def delete(id)
    begin
      zone = find_zone(id)
    rescue ZoneNotFound
      log.warn "######### DELETE #{id} SILENTLY IGNORED SINCE NOT FOUND ############"
      return nil
    end

    log.info "######### DELETE #{zone.name} ############"
    run_event_script zone, :before_delete

    zone.logging_on(plog) do
      zone.delete!
    end

    run_event_script zone, :after_delete
    log.info "######### DELETE #{zone.name} DONE #######"

    nil
  end

  operation :convert_database do
    help_caption ''
    help_text ''
  end
  def convert_database
    Zone.fix_database(request_data)
  end

  private

  def run_event_script(zone, event)
    return if agentcfg[:execute].blank? || agentcfg[:execute][event].blank?
    script = agentcfg[:execute][event].gsub('%z', zone.name)
    IO.popen(script, :err => [ :child, :out ]) { |io| plog.error io.read }
  end

end

end
end
end
