require 'fileutils'

require 'ygg/provisioner/model'

module Ygg
module Dns
module AuthoritativeProvisioner

class Zone < Ygg::Provisioner::Model

  self.state_persistant_keys = [ ]

  def update_config(serialized_model)
    super

    # Conversion
    zone_servers.each do |x|
      if x['server']['ipv4']
        x['server']['ipv4_addr_string'] = x['server']['ipv4']
        x['server'].delete 'ipv4'
      end

      if x['server']['ipv6']
        x['server']['ipv6_addr_string'] = x['server']['ipv6']
        x['server'].delete 'ipv6'
      end
    end

    server_entries = zone_servers.select { |x| x['server']['node_name'] == agentcfg[:nodename] }

    raise DataError, 'More than one server matches me' if server_entries.count > 1
    raise DataError, 'I\'m not listed as a zone server for this zone' if server_entries.blank?

    self.server_entry = server_entries.first

    log.debug "#{name}: This agent is responsible as server #{server_entry['server']['name']} - " +
              "IPv4=#{server_entry['server']['ipv4_addr_string']} - " +
              "IPv6=#{server_entry['server']['ipv6_addr_string']}"

    self.masters = zone_servers.select { |x| x['master'] }

    log.debug "Masters are [ #{masters.map { |x| x['server']['name'] }.join(', ')} ]"

    if master?
      log.debug "We are teh master!"
    else
      log.debug "We are a slave :("

      if masters.empty?
        raise DataError, 'Zone is configured as slave but masters list is empty'
      end
    end
  end

  def has_extra?
    File.exists?("/etc/named.d/extras/#{self.name}.conf")
  end

  def master?
    server_entry['master']
  end

  def zone_file
    if cfg[:deep_zones_dir]
      File.join(cfg[:zones_dir], *name.split('.').reverse) + '.zone'
    else
      File.join(cfg[:zones_dir], name + '.zon')
    end
  end

  def zone_file_new
    zone_file + '.new'
  end

  def prepare!
    # Regenerate & check named.conf
    self.class.generate_namedconf_file

    self.class.rndc_checkconf(plog)

    # Generate  & check zone file
    if master?
      generate_zone_file
      rndc_checkzone
    else
      check_slave_zone
    end
  end

  def activate!
    if master?
      switch_files(zone_file_new, zone_file)
    end
    
    self.class.activate_named_conf!
    self.class.rndc_reload(plog)
  end
  
  def create!
    prepare!
    activate!
    thaw! if has_extra?
  end
  
  def update!
    freeze! if has_extra?
    prepare!
    activate!
    thaw! if has_extra?
  end

  def freeze!
    if agentcfg[:rndc_bin]
      log.debug " -> freeze zone #{name}"
      self.class.run_command([ agentcfg[:rndc_bin], '-c', agentcfg[:rndc_conf], 'freeze', name ])
    end
  end

  def thaw!
    if agentcfg[:rndc_bin]
      log.debug " -> thaw zone #{name}"
      self.class.run_command([ agentcfg[:rndc_bin], '-c', agentcfg[:rndc_conf], 'thaw', name ])
    end
  end

  def delete!
    ######## Preparation

    # Regenerate & check named.conf
    self.class.delete(id)
    self.class.generate_namedconf_file
    self.class.rndc_checkconf(plog)

    ########

    self.class.activate_named_conf!
    self.class.rndc_reload(plog)

    begin
      File.unlink(zone_file)
    rescue Errno::ENOENT
    end
  end

  def record_is_comment?(r)
    r[:rr_class] == 'META' && r[:rr_type] == 'COMMENT'
  end

  def record_is_metadata(r)
    r[:rr_class] == 'META'
  end

  def multiline_comment(comment)
    comment.each_line.inject('') {|out, line| out << ";" + line; out}
  end

  def build_record(r)
    if record_is_comment?(r)
      multiline_comment(r[:comment]) + "\n"
    elsif record_is_metadata(r)
      ''
    else
      %Q[#{r[:name] || '@'}\t#{r[:ttl] || ''}\t#{r[:rr_class]}\t#{r[:rr_type]}\t#{r[:data]}#{r[:comment] ? "\t" + multiline_comment(r[:comment]) : ''}\n]
    end
  end

  def check_slave_zone(filename = zone_file)
    verify_zone_path filename
  end

  def generate_zone_file(filename = zone_file_new)
    verify_zone_path filename

    File::open(filename, 'w') do |f|
      f << %Q[; WARNING: This zone file has been autogenerated\n]
      f << %Q[;\n]
      f << %Q[; Zone: #{name}\n]
      f << %Q[; Generated: #{Time.now}\n]
      f << %Q[;\n\n]

      f << %Q[$TTL #{default_ttl}\n\n]

      records.sort! { |a,b|
        if a[:rr_class] == 'IN' && a[:rr_type] == 'SOA'
          -1
        elsif b[:rr_class] == 'IN' && b[:rr_type] == 'SOA'
          1
        else
          (a[:linenumber] || 1000000) <=> (b[:linenumber] || 1000000)
        end
      }

      if records[0][:rr_class] != 'IN' || records[0][:rr_type] != 'SOA'
        raise "Zone is missing SOA record"
      end

      records.each do |r|
        f << build_record(r)
      end

      if additional_text
        f << %Q[; \n]
        f << %Q[; Additional text \n]
        f << %Q[; \n]
        f << additional_text
      end
    end
  end

  def rndc_checkzone(filename = zone_file_new)
    run_command([ agentcfg[:named_checkzone_path], '-c', 'IN', name, filename ], stderr_handler: ->(out){ plog.error " -> #{out}"})
  rescue CommandError => e
    log.error " -> " + e.to_s
    e.backtrace.each {| b | log.error " ->    " + b }
    raise NamedZoneCheckFailed
  end

  def self.generate_namedconf_file(filename = agentcfg[:generated_named_conf_path] + '.new')

    File::open(filename, 'w') do |f|
      f << %Q[// WARNING: This file has been autogenerated, you should not edit it!!!!\n]
      f << %Q[//\n]
      f << %Q[// Generated: #{Time.now}\n]
      f << %Q[//\n\n]

      each do |zone|
        if zone.master?
          if zone.has_extra?
            f << %Q[zone "#{zone.name}" { type master; file "#{zone.zone_file}"; include "/etc/named.d/extras/#{zone.name}.conf"; };\n]
          else
            f << %Q[zone "#{zone.name}" { type master; file "#{zone.zone_file}"; };\n]
          end
        else
          master_ips = zone.masters.map { |x|
            [ x['server']['ipv4_addr_string'], x['server']['ipv6_addr_string'] ] }.flatten.reject { |ip| ip.blank? }

          f << %Q[zone "#{zone.name}" { type slave; file "#{zone.zone_file}"; masters { #{master_ips.join('; ')}; }; };\n]
        end
      end
    end
  end

  def self.activate_named_conf!
    switch_files(agentcfg[:generated_named_conf_path] + '.new', agentcfg[:generated_named_conf_path])
  end

  class NamedConfCheckFailed < Ygg::Provisioner::Task::PermanentFailure
    def initialize(msg  = 'Error checking bind named.conf file') ; super ; end
  end

  def self.rndc_checkconf(plog, conf_file = agentcfg[:generated_named_conf_path] + '.new')
    run_command([ agentcfg[:named_checkconf_path], conf_file ], plog: plog, stderr_handler: ->(out){ plog.error " -> #{out}"})
  rescue CommandError => e
    log.error " -> " + e.to_s
    e.backtrace.each {| b | log.error " ->    " + b }
    raise NamedConfCheckFailed
  end

  class NamedZoneCheckFailed < Ygg::Provisioner::Task::PermanentFailure
    def initialize(msg  = 'Error checking bind zone file') ; super ; end
  end

  class NamedReloadFailed < Ygg::Provisioner::Task::TemporaryFailure
    def initialize(msg  = 'Error reloading bind') ; super ; end
  end

  def self.rndc_reload(plog)
    if agentcfg[:rndc_bin]
      run_command([ agentcfg[:rndc_bin], '-c', agentcfg[:rndc_conf], 'reload' ], plog: plog, stderr_handler: ->(out){ plog.error " -> #{out}"})
    end
  rescue CommandError => e
    log.error " -> " + e.to_s
    e.backtrace.each {| b | log.error " ->    " + b }
    raise NamedReloadFailed
  end

  def verify_zone_path(filename)
    path = File.dirname(filename)
    if !FileTest::directory?(path)
      FileUtils.mkdir_p(path)
    end
  end
end

end
end
end
